@@ -6,19 +6,16 @@ Created on Fri Jun 10 21:00:21 2016
 """
 
 import os
-import logging
 import sys
+import logging
 import runpy
-import shutil
 import tempfile
 import pytest
 from typing import Dict
 
-import xarray as xr
 import numpy as np
-import pandas as pd
 import orca
-from wiver.wiver_python import (WIVER,
+from wiver.wiver_python import (WIVER, InvalidWiverInputData,
                                 DestinationChoiceError, DataConsistencyError)
 import wiver.run_wiver
 from netCDF4 import __hdf5libversion__
@@ -40,44 +37,35 @@ def param_dist(request) -> float:
     return request.param
 
 
-@pytest.fixture(scope='class', params=[1.2, 10])
-def param_saving_weights(request):
+@pytest.fixture(scope='class', params=[1.01, 1.1, 1.5, 2, np.inf])
+def param_savings_param(request):
     """
-    different values for saving categories
+    different values for savings parameter
     """
     return request.param
 
 
-@pytest.fixture(scope='class', params=range(2))
+@pytest.fixture(scope='class', params=range(5))
 def zone_h(request) -> int:
     """
     different values for home zone
     """
     return request.param
 
-@pytest.fixture(scope='class',
+@pytest.fixture(scope='module')
+def folder(tmpdir_factory) -> str:
+    """return a tempdir managed by pytest"""
+    return tmpdir_factory.mktemp('Wiver').strpath
+
+
+@pytest.fixture(scope='module',
                 params=['Wiver',
                         pytest.param('ÄÖÜß', marks=pytest.mark.xfail(
                                      __hdf5libversion__ < '1.12',
                             reason='Bug in HDF5'))])
-def folder(request) -> str:
-    """
-    temp folder
-    tempfolder with Umlaut causes problems because of HDF5-Bug,
-    so use a folder without Umlaut until bug will be fixed in hdf5 >= 1.12
-    """
-    folder = tempfile.mkdtemp(prefix=request.param)
-    yield folder
-    def handleLockedLogfileError(func, path, exc_info):
-        logging.shutdown()
-        func(path)
-    shutil.rmtree(folder, onerror=handleLockedLogfileError)
-
-
-@pytest.fixture(scope='class')
-def result_folder(folder: str) -> str:
+def result_folder(request, folder: str) -> str:
     """Temp folder to store the results"""
-    result_folder = os.path.join(folder, 'results')
+    result_folder = os.path.join(folder, f'results_{request.param}')
     os.makedirs(result_folder, exist_ok=True)
     return result_folder
 
@@ -144,33 +132,26 @@ def wiver(request) -> WIVER:
     wiver = WIVER(n_groups=3,
                   n_zones=5,
                   n_time_slices=3,
-                  n_savings_categories=10,
                   n_modes=3,
                   n_sectors=len(sector_name))
 
     # mode of group
     wiver.mode_g = np.array([2, 2, 1])
     # define centroids of zones
-    x = np.array([2, 5, 1, 6, 10])
-    y = np.array([5, 2, 3, 1, 7])
+    # 2 zones at the same spot
+    x = np.array([2, 1, 1, 6, 10])
+    y = np.array([5, 3, 3, 1, 7])
     # calc travel time between zones
     dx = x - x[:, np.newaxis]
     dy = y - y[:, np.newaxis]
     dist = np.sqrt(dx**2 + dy**2)
     wiver.km_ij = dist
-    t = dist * 3 + 1
+    t = dist * 3
 
     # travel time is identical for all modes
     wiver.travel_time_mij[:] = t
-    # define savings intervals
-    # the first interval represents NINF (no connection)
-    # for this interval the saving weight is 0, wich leads to 'no connection'
-    wiver.savings_bins_s = np.concatenate(([np.NINF],
-                                           np.linspace(.15, .85, 8),
-                                           [1]))
-    wiver.savings_weights_gs[0] = np.linspace(0, 2, wiver.n_savings_categories)
-    wiver.savings_weights_gs[1] = np.linspace(0, 5, wiver.n_savings_categories)
-    wiver.savings_weights_gs[2] = np.linspace(0, 10, wiver.n_savings_categories)
+    # define savings parameters
+    wiver.savings_param_g = [2, 1.1, 1.01]
 
     wiver.source_potential_gh = np.array([[10, 20, 10, 0, 10],
                                          [100, 20, 10, 0, 0],
@@ -204,7 +185,7 @@ def wiver(request) -> WIVER:
     wiver.mode_name = np.array(['Rad', 'Pkw', 'OV'])
     wiver.modes = np.array(['R', 'P', 'O'])
     wiver.groups = np.array([200, 201, 999])
-    wiver.group_names = np.array(['Gruppe {}'.format(i) for i in wiver.groups])
+    wiver.group_names = np.array([f'Gruppe {i}' for i in wiver.groups])
 
     return wiver
 
@@ -232,7 +213,6 @@ class Test01_WiverData:
         assert dims['destinations'] == wiver.n_zones
         assert dims['modes'] == wiver.n_modes
         assert dims['groups'] == wiver.n_groups
-        assert dims['savings'] == wiver.n_savings_categories
         assert dims['time_slices'] == wiver.n_time_slices
 
     def test_03_save_data(self, wiver: WIVER, wiver_files: Dict[str, str]):
@@ -276,6 +256,15 @@ class Test01_WiverData:
         wiver.save_results_to_visum(folder, visum_format='V')
         wiver.save_results_to_visum(folder, visum_format='BK')
 
+    def test_07_validate_data(self, wiver: WIVER):
+        """Test the validation"""
+        wiver.stops_per_tour_g[:] = [1, 2, 3]
+        wiver.validate_input_data
+        # no values < 1 allowed
+        wiver.stops_per_tour_g[:] = [0.8, 2, 3]
+        with pytest.raises(InvalidWiverInputData) as e:
+            wiver.validate_input_data()
+
 
 class Test02_Wiver:
     """Test Wiver Model"""
@@ -296,14 +285,22 @@ class Test02_Wiver:
         h = 0
         t = 0
         wiver.calc_destination_choice(t, group, h)
-        before = wiver.p_destination_tij[t, h].copy()
+        before = wiver.p_destination_tj[t].copy()
         print(before)
 
         wiver.param_dist_g[group] = param_dist
         wiver.calc_destination_choice(t, group, h)
-        after = wiver.p_destination_tij[t, h].copy()
+        after = wiver.p_destination_tj[t].copy()
         print(after)
 
+        m = wiver.mode_g[group]
+        res = np.zeros((wiver.n_zones))
+        for j in range(wiver.n_zones):
+            res[j] = wiver.calc_p_destination(group, m, h, j)
+        res /= res.sum()
+        print(res)
+        np.testing.assert_allclose(res, wiver.p_destination_tj[t])
+
     def test_03_savings(self, wiver: WIVER, zone_h: int):
         """Test the savings function for closeby points 1 and 3
         from different home zones (parameter: zone_h)"""
@@ -319,12 +316,26 @@ class Test02_Wiver:
         print(s)
         assert s == s_target
 
-    def test_04_savings_factor(self, wiver: WIVER, zone_h: int):
+        # edge case where t_hi and t_jh are 0
+        wiver.travel_time_mij[m, zone_h, i] = 0
+        wiver.travel_time_mij[m, j, zone_h] = 0
+        wiver.travel_time_mij[m, i, j] = 0
+        s = wiver.calc_savings(g, m, zone_h, i, j)
+        assert s == 0
+        #  linking trip is a detour and longer than the trip back to the depot
+        wiver.travel_time_mij[m, i, j] = 1
+        s = wiver.calc_savings(g, m, zone_h, i, j)
+        assert s == -99
+
+    def test_04_savings_factor(self, wiver: WIVER,
+                               zone_h: int,
+                               param_savings_param: float):
         """Test the savings factor function"""
         g = 0
         i = 1
         j = 3
         m = wiver.mode_g[g]
+        wiver.savings_param_g[g] = param_savings_param
         sf = wiver.calc_savings_factor(g, m, zone_h, i, j)
         print(sf)
 
@@ -367,31 +378,31 @@ class Test02_Wiver:
             zone_interior_link_trips = link_matrix.diagonal().sum()
             return zone_interior_link_trips
 
-        h = 0
+        h = 2
         t = 0
-        group = 1
+        group = 0
         tours = 100
         linking_trips = 200
-        n_sav = wiver.n_savings_categories
+        wiver.param_dist_g[group] = -0.02
         wiver.calc_destination_choice(t, group, h)
 
         # test first without any savings
-        wiver.savings_weights_gs[group] = 1
+        wiver.savings_param_g[group] = np.inf
         interior_trips_no_savings = check_trips_with_savings(
             wiver, group, h, t, tours, linking_trips)
 
         # test first with low savings
-        wiver.savings_weights_gs[group] = np.linspace(1, 2, num=n_sav)
+        wiver.savings_param_g[group] = 2
         interior_trips_low_savings = check_trips_with_savings(
             wiver, group, h, t, tours, linking_trips)
 
         # test first with medium savings
-        wiver.savings_weights_gs[group] = np.linspace(1, 5, num=n_sav)
+        wiver.savings_param_g[group] = 1.1
         interior_trips_medium_savings = check_trips_with_savings(
             wiver, group, h, t, tours, linking_trips)
 
         # test first with strong savings
-        wiver.savings_weights_gs[group] = np.linspace(1, 10, num=n_sav)
+        wiver.savings_param_g[group] = 1.01
         interior_trips_high_savings = check_trips_with_savings(
             wiver, group, h, t, tours, linking_trips)
 
@@ -406,7 +417,14 @@ class Test02_Wiver:
 
     def test_08_calc_time_series(self, wiver: WIVER):
         """Test the time_series"""
-        wiver.calc_daily_trips()
+        for g in range(wiver.n_groups):
+            wiver.calc_daily_trips(g)
+            wiver.adjust_linking_trips(g)
+            wiver.trips_gij[g] = (wiver.home_based_trips_gij[g] +
+                                wiver.linking_trips_gij[g] +
+                                wiver.return_trips_gij[g])
+            wiver.trips_to_destination_gj[g] = (wiver.home_based_trips_gij[g] +
+                                               wiver.linking_trips_gij[g]).sum(0)
         wiver.calc_time_series()
 
         # test the normalisation of the time series
@@ -425,15 +443,35 @@ class Test02_Wiver:
     def test_09_trips_gij(self, wiver: WIVER):
         """Test if the arrays are not overwritten"""
         arr_gij_before = wiver.trips_gij.__array_interface__
-        wiver.calc_daily_trips()
+        for g in range(wiver.n_groups):
+            wiver.calc_daily_trips(g)
         arr_gij_after = wiver.trips_gij.__array_interface__
         assert arr_gij_before == arr_gij_after
 
     def test_10_wiver_calc(self, wiver: WIVER):
-        """Test the whole model"""
+        """Test the whole model with different number of threads"""
         wiver.calc()
+        res = wiver.trips_gsij
         print(wiver.trips_gsij.sum(-1).sum(-1))
 
+        wiver.set_n_threads(1)
+        wiver.calc()
+        res2 = wiver.trips_gsij
+        np.testing.assert_allclose(res, res2)
+
+        wiver.set_n_threads(16)
+        wiver.calc()
+        res3 = wiver.trips_gsij
+        np.testing.assert_allclose(res, res3)
+
+        # test if group without linking trips (only one stop) works
+        g = 2
+        wiver.stops_per_tour_g[g] = 1
+        wiver.calc()
+        res = wiver.trips_gij[g]
+        assert np.isfinite(res).all()
+
+
     def test_21_no_destinations(self, wiver: WIVER):
         """Test exceptions when there is demand
         but no destinations accessible"""
@@ -444,7 +482,7 @@ class Test02_Wiver:
         t = 0
         group = 0
         wiver.calc_destination_choice(t, group, h)
-        print(wiver.p_destination_tij[t, h])
+        print(wiver.p_destination_tj[t])
 
         # now set sink_potential for group 0 to 0
         wiver.sink_potential_gj[0] = 0
@@ -506,7 +544,6 @@ class Test02_Wiver:
         wiver.calc()
 
         g = 1
-        h = 0
         m = wiver.mode_g[g]
         # infinitive travel times from home zone 0
         backup = wiver.travel_time_mij[m].copy()
@@ -559,13 +596,12 @@ class Test02_Wiver:
         after = wiver.mean_distance_g
         print(after)
 
-    def test_41_mean_distance_savings(self, wiver: WIVER, param_saving_weights: float):
+    def test_41_mean_distance_savings(self, wiver: WIVER, param_savings_param: float):
         """
         Test the mean distance calculation
         with different savings optimizations
         """
-        wiver.savings_weights_gs[0] = np.linspace(1, param_saving_weights,
-                                                  wiver.n_savings_categories)
+        wiver.savings_param_g[0] = param_savings_param
         wiver.calc()
         after = wiver.mean_distance_g
         print(after)
@@ -588,17 +624,18 @@ class Test02_Wiver:
         wiver.define_datasets()
         sp = wiver.zonal_data.sink_potential
         target_share = sp / sp.sum('destinations')
-        wiver.calc_with_balancing(max_iterations=1)
-        target_total_trips = wiver.trips_gij.sum()
-        wiver.calc_with_balancing(max_iterations=100)
+        wiver.calc_with_balancing(max_iterations=100, threshold=0.001)
         trips = wiver.balancing.trips_to_destination
         actual_share = trips / trips.sum('destinations')
         np.testing.assert_allclose(actual_share, target_share, rtol=.1)
 
+        target_total_trips = wiver.trips_gij.sum()
         actual_total_trips = wiver.trips_gij.sum()
         np.testing.assert_allclose(actual_total_trips,
                                    target_total_trips,
                                    rtol=.01)
+        for g, group in enumerate(wiver.groups):
+            print(f'Group {group} converged after {wiver.converged_g[g]} iterations.')
 
     def test_51_test_starting_equals_ending_trips(self, wiver: WIVER):
         """Test that starting and ending trips per zone are equal"""
@@ -665,7 +702,9 @@ class Test03_TestExport:
 
             orca.run(steps)
 
-    def test_22_run_wiver(self, folder: str, wiver: WIVER, wiver_files: Dict[str, str]):
+    def test_22_run_wiver(self, folder: str,
+                          wiver: WIVER,
+                          wiver_files: Dict[str, str]):
         """Test wiver.run_wiver with command line parameters"""
         backup_sys_argv = sys.argv
         os.makedirs(folder, exist_ok=True)
